@@ -49,24 +49,7 @@
 #include <unistd.h>
 #include <string.h>
 
-#ifdef HAVE_LIBSSL
-#include <openssl/ssl.h>
-#include <openssl/err.h>
-
-#define SRV_SSL_V23 0
-#define SRV_SSL_V2 1
-#define SRV_SSL_V3 2
-#define SRV_SSL_TLS1 3
-
-static char ssl_compat;
-static char require_peer_cert;
-static char ssl_protocol;
-static char *certfile;
-static char *keyfile;
-static char *cacert_dir;
-static char *cacert_file;
-static SSL_CTX *ssl_context = NULL;
-#endif  /* HAVE_LIBSSL */
+#include "ssl.h"
 
 #ifdef HAVE_LIBGEOIP
 #include <GeoIP.h>
@@ -85,38 +68,13 @@ GeoIP *geoip4, *geoip6;
 #define CLIENTS_MAX	2048	/* max clients */
 #define SERVERS_MAX	16	/* max servers */
 #define ACLS_MAX	10	/* max acls */
-#define CONNECTIONS_MAX	256	/* max simultaneous connections */
+#define CONNECTIONS_MAX	500	/* max simultaneous connections */
 #define TIMEOUT		3	/* default timeout for non reachable hosts */
 #define BLACKLIST_TIME	30	/* how long to shun a server that is down */
 #define TRACKING_TIME	0	/* how long a client is remembered */
 #define KEEP_MAX	100	/* how much to keep from the URI */
 #define WEIGHT_FACTOR	256	/* to make weight kick in earlier */
 
-/* Connection states, as used in struct conn */
-#define CS_UNUSED	(0)	/* not connected, unused slot */
-#define CS_IN_PROGRESS	(1)	/* connection in progress */
-#define CS_CONNECTED	(2)	/* successfully connected */
-#define CS_CLOSED_UP	(4)	/* we read eof from upfd */
-#define CS_CLOSED_DOWN	(8)	/* we read eof from downfd */
-#define CS_CLOSED	(CS_CLOSED_UP | CS_CLOSED_DOWN)
-
-typedef struct {
-	int state;		/* as per above */
-	time_t t;		/* time of connection attempt */
-	int downfd, upfd;
-	unsigned char *downb, *downbptr, *upb, *upbptr;
-	int downn, upn;		/* pending bytes */
-	unsigned long ssx, srx;	/* server total sent, received */
-	unsigned long csx, crx;	/* client total sent, received */
-	int client;		/* client index */
-	int initial;		/* first server tried */
-	int server;		/* server index */
-	int pend;		/* node in pending_list */
-#ifdef HAVE_LIBSSL
-	SSL *ssl;
-#endif
-} connection;
-
 typedef struct {
 	int status;		/* last failed connection attempt */
 	int acl;		/* which clients can use this server */
@@ -181,11 +139,13 @@ static int idle_timeout = 0;	/* never time out */
 static int abort_on_error = 0;
 static int dummy = 0;		/* use pen as a test target */
 static int idlers = 0, idlers_wanted = 0;
-static time_t now;
+time_t now;
 static int tcp_nodelay = 0;
 static int tcp_fastclose = 0;
 static int pending_list = -1;	/* pending connections */
-static int listen_queue = 100;
+static int pending_queue = 0;	/* number of pending connections */
+static int pending_max = 100;	/* max number of pending connections */
+static int listen_queue = CONNECTIONS_MAX;
 static int multi_accept = 100;
 static int nservers;		/* number of servers */
 static int current;		/* current server */
@@ -227,8 +187,6 @@ static int client_acl, control_acl;
 static int udp = 0;
 static int protoid = SOCK_STREAM;
 
-static int port;
-
 static char *cfgfile = NULL;
 static char *logfile = NULL;
 static FILE *logfp = NULL;
@@ -238,6 +196,8 @@ static char *pidfile = NULL;
 static FILE *pidfp = NULL;
 static char *webfile = NULL;
 static char listenport[1000];
+static int port;
+
 static char *ctrlport = NULL;
 static int listenfd, ctrlfd = -1;
 static char *e_server = NULL;
@@ -486,7 +446,7 @@ static int accept_nb(int sockfd, struct sockaddr *addr, socklen_t *addrlen)
 	if (tcp_nodelay) tcp_nodelay_on(sockfd);
 }
 
-static struct sigaction alrmaction, hupaction, termaction, usr1action;
+static struct sigaction alrmaction, hupaction, termaction, usr1action, usr2action;
 
 void debug(char *fmt, ...)
 {
@@ -548,7 +508,7 @@ void *pen_realloc(void *p, size_t n)
 	return q;
 }
 
-static char *pen_strdup(char *p)
+char *pen_strdup(const char *p)
 {
 	size_t len = strlen(p);
 	char *b = pen_malloc(len+1);
@@ -830,156 +790,6 @@ static int pen_aton(char *name, struct sockaddr_storage *addr)
 	return result;
 }
 
-#ifdef HAVE_LIBSSL
-static int ssl_verify_cb(int ok, X509_STORE_CTX *ctx)
-{
-	char buffer[256];
-
-	X509_NAME_oneline(X509_get_issuer_name(ctx->current_cert),
-			buffer, sizeof(buffer));
-	if (ok) {
-		debug("SSL: Certificate OK: %s", buffer);
-	} else {
-		switch (ctx->error) {
-		case X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT:
-			debug("SSL: Cert error: CA not known: %s", buffer);
-			break;
-		case X509_V_ERR_CERT_NOT_YET_VALID:
-			debug("SSL: Cert error: Cert not yet valid: %s",
-				buffer);
-			break;
-		case X509_V_ERR_ERROR_IN_CERT_NOT_BEFORE_FIELD:
-			debug("SSL: Cert error: illegal \'not before\' field: %s",
-				buffer);
-			break;
-		case X509_V_ERR_CERT_HAS_EXPIRED:
-			debug("SSL: Cert error: Cert expired: %s", buffer);
-			break;
-		case X509_V_ERR_ERROR_IN_CERT_NOT_AFTER_FIELD:
-			debug("SSL: Cert error: invalid \'not after\' field: %s",
-				buffer);
-			break;
-		default:
-			debug("SSL: Cert error: unknown error %d in %s",
-				ctx->error, buffer);
-			break;
-		}
-	}
-	return ok;
-}
-
-static RSA *ssl_temp_rsa_cb(SSL *ssl, int export, int keylength)
-{
-	static RSA *rsa = NULL;
-
-	if (rsa == NULL)
-		rsa = RSA_generate_key(512, RSA_F4, NULL, NULL);
-	return rsa;
-}
-
-static int ssl_init(void)
-{
-	int err;
-
-	SSL_load_error_strings();
-	SSLeay_add_ssl_algorithms();
-	switch (ssl_protocol) {
-#if 0
-	case SRV_SSL_V2:
-		ssl_context = SSL_CTX_new(SSLv2_method());
-		break;
-#endif
-	case SRV_SSL_V3:
-		ssl_context = SSL_CTX_new(SSLv3_method());
-		break;
-	default:
-	case SRV_SSL_V23:
-		ssl_context = SSL_CTX_new(SSLv23_method());
-		break;
-	case SRV_SSL_TLS1:
-		ssl_context = SSL_CTX_new(TLSv1_method());
-		break;
-	}
-	if (ssl_context == NULL) {
-		err = ERR_get_error();
-		error("SSL: Error allocating context: %s",
-			ERR_error_string(err, NULL));
-	}
-	if (ssl_compat) {
-		SSL_CTX_set_options(ssl_context, SSL_OP_ALL);
-	}
-	if (certfile == NULL || *certfile == 0) {
-		debug("SSL: No cert file specified in config file!");
-		error("The server MUST have a certificate!");
-	}
-	if (keyfile == NULL || *keyfile == 0)
-		keyfile = certfile;
-	if (certfile != NULL && *certfile != 0) {
-		if (!SSL_CTX_use_certificate_file(ssl_context, certfile,
-						SSL_FILETYPE_PEM)) {
-			err = ERR_get_error();
-			error("SSL: error reading certificate from file %s: %s",
-				certfile, ERR_error_string(err, NULL));
-		}
-		if (!SSL_CTX_use_PrivateKey_file(ssl_context, keyfile,
-						SSL_FILETYPE_PEM)) {
-			err = ERR_get_error();
-			error("SSL: error reading private key from file %s: %s",
-				keyfile, ERR_error_string(err, NULL));
-		}
-		if (!SSL_CTX_check_private_key(ssl_context)) {
-			error("SSL: Private key does not match public key in cert!");
-		}
-	}
-	if (cacert_dir != NULL && *cacert_dir == 0)
-		cacert_dir = NULL;
-	if (cacert_file != NULL && *cacert_file == 0)
-		cacert_file = NULL;
-	if (cacert_dir != NULL || cacert_file != NULL) {
-		if (!SSL_CTX_load_verify_locations(ssl_context,
-					cacert_file, cacert_dir)) {
-			err = ERR_get_error();
-			debug("SSL: Error error setting CA cert locations: %s",
-				ERR_error_string(err, NULL));
-			cacert_file = cacert_dir = NULL;
-		}
-	}
-	if (cacert_dir == NULL && cacert_file == NULL) {  /* no verify locations loaded */
-		debug("SSL: No verify locations, trying default");
-		if (!SSL_CTX_set_default_verify_paths(ssl_context)) {
-			err = ERR_get_error();
-			debug("SSL: Error error setting default CA cert location: %s",
-				ERR_error_string(err, NULL));
-			debug("continuing anyway...");
-		}
-	}
-	SSL_CTX_set_tmp_rsa_callback(ssl_context, ssl_temp_rsa_cb);
-	if (require_peer_cert) {
-		SSL_CTX_set_verify(ssl_context,
-			SSL_VERIFY_PEER | SSL_VERIFY_FAIL_IF_NO_PEER_CERT,
-			ssl_verify_cb);
-	} else {
-		SSL_CTX_set_verify(ssl_context,
-			SSL_VERIFY_NONE,
-			ssl_verify_cb);
-	}
-
-	SSL_CTX_set_client_CA_list(ssl_context,
-			SSL_load_client_CA_file(certfile));
-
-	/* permit large writes to be split up in several records */
-	SSL_CTX_set_mode(ssl_context, SSL_MODE_ENABLE_PARTIAL_WRITE);
-
-#if 1	/* testing */
-	debug("SSL_CTX_get_session_cache_mode() returns %d",
-		SSL_CTX_get_session_cache_mode(ssl_context));
-	SSL_CTX_set_session_cache_mode(ssl_context, 0);
-#endif
-
-	return 0;
-}
-
-#endif  /* HAVE_LIBSSL */
 
 /* allocate ace and fill in the generics */
 static int add_acl(int a, unsigned char permit)
@@ -1398,6 +1208,11 @@ static void quit(int dummy)
 	loopflag = 0;
 }
 
+static void die(int dummy)
+{
+	abort();
+}
+
 /* Store client and return index */
 static int store_client(struct sockaddr_storage *cli)
 {
@@ -1873,20 +1688,10 @@ static int copy_down(int i)
 
 #ifdef HAVE_LIBSSL
 		if (ssl) {
-			/* can't write more than 32000 bytes at a time */
-			int ssl_rc;
-			if (rc > 32000) ssl_rc = 32000;
-			else ssl_rc = rc;
-			n = SSL_write(ssl, b, ssl_rc);
-			DEBUG(2, "SSL_write returns %d", n);
-			if (n < 0) {
-				err = SSL_get_error(ssl, n);
-				if (debuglevel) debug("SSL error %d\n", err);
-				if (err == SSL_ERROR_WANT_READ ||
-				    err == SSL_ERROR_WANT_WRITE) {
-					return 0;
-				}
-			}
+			/* We can't SSL_write here, because the auto buffer we're using now
+			   won't be around if we need to retry the write.
+			   Therefore don't write anything here but let flush_down do it. */
+			n = 0;
 		} else {
 			n = my_send(to, b, rc, 0);
 			err = socket_errno;
@@ -1990,6 +1795,23 @@ static void close_conn(int i)
 		close(conns[i].upfd);
 		fd2conn_set(conns[i].upfd, -1);
 	}
+#ifdef HAVE_LIBSSL
+	if (conns[i].ssl) {
+		int n = SSL_shutdown(conns[i].ssl);
+		DEBUG(3, "First SSL_shutdown(%d) returns %d", conns[i].ssl, n);
+		if (n == 0) {
+			n = SSL_shutdown(conns[i].ssl);
+			DEBUG(3, "Second SSL_shutdown(%d) returns %d", conns[i].ssl, n);
+		}
+		if (n == -1) {
+			n = SSL_get_error(conns[i].ssl, n);
+			DEBUG(3, "%s", ERR_error_string(SSL_get_error(conns[i].ssl, n), NULL));
+			ERR_print_errors_fp(stderr);
+		}
+		SSL_free(conns[i].ssl);
+		conns[i].ssl = 0;
+	}
+#endif
 	if (conns[i].downfd != -1 && conns[i].downfd != listenfd) {
 		event_delete(conns[i].downfd);
 		close(conns[i].downfd);
@@ -2005,12 +1827,6 @@ static void close_conn(int i)
 		free(conns[i].upb);
 		conns[i].upn=0;
 	}
-#ifdef HAVE_LIBSSL
-	if (conns[i].ssl) {
-		SSL_free(conns[i].ssl);
-		conns[i].ssl = 0;
-	}
-#endif
 	connections_used--;
 	DEBUG(2, "decrementing connections_used to %d for connection %d",
 		connections_used, i);
@@ -2020,6 +1836,7 @@ static void close_conn(int i)
 	}
 	if (conns[i].state == CS_IN_PROGRESS) {
 		pending_list = dlist_remove(conns[i].pend);
+		pending_queue--;
 	}
 	conns[i].state = CS_UNUSED;
 	DEBUG(2, "close_conn: Closing connection %d to server %d; connections_used = %d\n" \
@@ -2278,6 +2095,7 @@ static int try_server(int index, int conn)
 		conns[conn].state = CS_IN_PROGRESS;
 		pending_list = dlist_insert(pending_list, conn);
 		conns[conn].pend = pending_list;
+		pending_queue++;
 		event_add(upfd, EVENT_WRITE);
 		DEBUG(2, "Pending connect to server %d\n" \
 			"conns[%d].client = %d\n" \
@@ -2288,6 +2106,7 @@ static int try_server(int index, int conn)
 			debug("Server %d failed, retry in %d sec: %d",
 				index, blacklist_time, socket_errno);
 		}
+		debug("blacklisting server %d because connect error %d", index, err);
 		blacklist_server(index);
 		close(upfd);
 		return 0;
@@ -2361,7 +2180,9 @@ static int open_listener(char *a)
 	char b[1024], *p;
 	int one = 1;
 	int optval = 1;
+#if 0
 	int port;
+#endif
 
 #ifndef WINDOWS
 	/* Handle Unix domain sockets separately */
@@ -2739,6 +2560,10 @@ static void do_cmd(char *b, void (*output)(void *, char *, ...), void *op)
 		} else if (!strcmp(p, "weight")) {
 			weight = 0;
 		}
+	} else if (!strcmp(p, "pending_max")) {
+		p = strtok(NULL, " ");
+		if (p) pending_max = atoi(p);
+		if (pending_max <= 0) pending_max = 1;
 	} else if (!strcmp(p, "pid")) {
 		output(op, "%ld\n", (long)getpid());
 	} else if (!strcmp(p, "poll")) {
@@ -2820,6 +2645,53 @@ static void do_cmd(char *b, void (*output)(void *, char *, ...), void *op)
 		} else {
 			output(op, "Unable to create webstats\n");
 		}
+#ifdef HAVE_LIBSSL
+	} else if (!strcmp(p, "ssl_ciphers")) {
+		p = strtok(NULL, " ");
+		if (ssl_ciphers) {
+			free(ssl_ciphers);
+			ssl_ciphers = NULL;
+		}
+		if (p) ssl_ciphers = pen_strdup(p);
+	} else if (!strcmp(p, "ssl_client_renegotiation_interval")) {
+		p = strtok(NULL, " ");
+		ssl_client_renegotiation_interval = atoi(p);
+	} else if (!strcmp(p, "ssl_ocsp_response")) {
+		p = strtok(NULL, " ");
+		if (ocsp_resp_file) {
+			free(ocsp_resp_file);
+		}
+		ocsp_resp_file = pen_strdup(p);
+	} else if (!strcmp(p, "ssl_option")) {
+		p = strtok(NULL, " ");
+		if (p == NULL) {
+			debug("Missing option");
+		} else if (!strcmp(p, "no_sslv2")) {
+			ssl_options |= SSL_OP_NO_SSLv2;
+		} else if (!strcmp(p, "no_sslv3")) {
+			ssl_options |= SSL_OP_NO_SSLv3;
+		} else if (!strcmp(p, "no_tlsv1")) {
+			ssl_options |= SSL_OP_NO_TLSv1;
+#ifdef SSL_OP_NO_TLSv1_1
+		} else if (!strcmp(p, "no_tlsv1.1")) {
+			ssl_options |= SSL_OP_NO_TLSv1_1;
+#endif
+#ifdef SSL_OP_NO_TLSv1_2
+		} else if (!strcmp(p, "no_tlsv1.2")) {
+			ssl_options |= SSL_OP_NO_TLSv1_2;
+#endif
+		} else if (!strcmp(p, "cipher_server_preference")) {
+			ssl_options |= SSL_OP_CIPHER_SERVER_PREFERENCE;
+		}
+	} else if (!strcmp(p, "ssl_sni_path")) {
+		p = strtok(NULL, " ");
+		if (p == NULL) {
+			debug("Missing ssl_sni_path");
+		} else {
+			if (ssl_sni_path) free(ssl_sni_path);
+			ssl_sni_path = pen_strdup(p);
+		}
+#endif	/* HAVE_SSL */
 	} else if (!strcmp(p, "stubborn")) {
 		stubborn = 1;
 	} else if (!strcmp(p, "tcp_fastclose")) {
@@ -2867,7 +2739,8 @@ static void do_cmd(char *b, void (*output)(void *, char *, ...), void *op)
 			debug("write: no file");
 		}
 	} else {
-		output(op, "do_cmd: ignoring command starting with '%s'\n", p);
+		if (p[0] != '#')
+			output(op, "do_cmd: ignoring command starting with '%s'\n", p);
 	}
 }
 
@@ -3109,6 +2982,8 @@ static void add_client(int downfd, struct sockaddr_storage *cli_addr)
 
 #ifdef HAVE_LIBSSL
 	conn = store_conn(downfd, ssl, client);
+	conns[conn].reneg = 0;	/* never */
+	SSL_set_app_data(ssl, &conns[conn]);
 #else
 	conn = store_conn(downfd, client);
 #endif
@@ -3152,7 +3027,9 @@ static int flush_down(int i)
 	SSL *ssl = conns[i].ssl;
 
 	if (ssl) {
-		n = SSL_write(ssl, conns[i].downbptr, conns[i].downn);
+		int m = conns[i].downn;
+		if (m > 16000) m = 16000;
+		n = SSL_write(ssl, conns[i].downbptr, m);
 		DEBUG(2, "SSL_write returns %d\n", n);
 		if (n < 0) {
 			int err = SSL_get_error(ssl, n);
@@ -3263,6 +3140,11 @@ static void setup_signals(void)
 	usr1action.sa_flags = 0;
 	sigaction(SIGUSR1, &usr1action, NULL);
 
+	usr2action.sa_handler = die;
+	sigemptyset(&usr2action.sa_mask);
+	usr2action.sa_flags = 0;
+	sigaction(SIGUSR2, &usr2action, NULL);
+
 	hupaction.sa_handler = restart_log;
 	sigemptyset(&hupaction.sa_mask);
 	hupaction.sa_flags = 0;
@@ -3312,6 +3194,7 @@ static void check_listen_socket(void)
 	/* process tcp connection(s) */
 		for (i = 0; i < multi_accept; i++) {
 			if (connections_used >= connections_max) break;
+			if (pending_queue >= pending_max) break;
 			downfd = accept_nb(listenfd,
 				(struct sockaddr *)&cli_addr, &clilen);
 			if (downfd < 0) {
@@ -3360,6 +3243,7 @@ static void check_if_connected(int i)
 	}
 	if (result != 0) {
 		debug("Connect failed: %s", strerror(result));
+		debug("blacklisting server %d because of connect failure", conns[i].server);
 		blacklist_server(conns[i].server);
 		if (failover_server(i) == 0) {
 			//close_conn(i);
@@ -3369,6 +3253,7 @@ static void check_if_connected(int i)
 	DEBUG(2, "Connection %d completed", i);
 	if (conns[i].state == CS_IN_PROGRESS) {
 		pending_list = dlist_remove(conns[i].pend);
+		pending_queue--;
 	}
 	conns[i].state = CS_CONNECTED;
 	conns[i].t = now;
@@ -3390,6 +3275,7 @@ static void check_if_timeout(int i)
 		(int)now, i, i, conns[i].t, i, now-conns[i].t);
 	if (now-conns[i].t >= timeout) {
 		DEBUG(2, "Connection %d timed out", i);
+		debug("blacklisting server %d because of connect timeout", conns[i].server);
 		blacklist_server(conns[i].server);
 		if (failover_server(i) == 0) {
 			//close_conn(i);
@@ -3502,6 +3388,10 @@ static int handle_events(int *pending_close)
                                 if (!try_flush_up(conn)) closing = 1;
                         }
                 }
+		if (conns[conn].state == CS_CLOSED) {
+			DEBUG(2, "Connection %d was closed", conn);
+			closing = 1;
+		}
 		if (closing) {
 			pending_close[npc++] = conn;
 		}
@@ -3802,10 +3692,13 @@ static int options(int argc, char **argv)
 
 int main(int argc, char **argv)
 {
-	int i;
+	int i, n;
 	struct passwd *pwd = NULL;
 	struct rlimit r;
-	int n = options(argc, argv);
+
+	init_mask();
+
+	n = options(argc, argv);
 	argc -= n;
 	argv += n;
 
@@ -3814,6 +3707,9 @@ int main(int argc, char **argv)
 	start_winsock();
 #endif
 
+	init(argc, argv);
+	read_cfg(cfgfile);
+
 	getrlimit(RLIMIT_CORE, &r);
 	r.rlim_cur = r.rlim_max;
 	setrlimit(RLIMIT_CORE, &r);
@@ -3826,12 +3722,6 @@ int main(int argc, char **argv)
 	}
 #endif
 
-#ifdef HAVE_LIBSSL
-	if (certfile) {
-		ssl_init();
-	}
-#endif
-
 	/* we must open listeners before dropping privileges */
 	/* Control port */
 	if (ctrlport) {
@@ -3850,8 +3740,12 @@ int main(int argc, char **argv)
 
 	snprintf(listenport, sizeof listenport, "%s", argv[0]);
 	listenfd = open_listener(listenport);
-	init_mask();
-	init(argc, argv);
+
+#ifdef HAVE_LIBSSL
+	if (certfile) {
+		ssl_init();
+	}
+#endif
 
 	/* we must look up user id before chrooting */
 	if (user) {
@@ -3874,7 +3768,9 @@ int main(int argc, char **argv)
 			error("Can't setuid(%d)", (int)pwd->pw_uid);
 	}
 
+#if 0
 	read_cfg(cfgfile);
+#endif
 	open_log(logfile);
 	if (pidfile) {
 		pidfp = fopen(pidfile, "w");
